/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2018	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	Implement utility level procedure to	*/
/*	local application specific configuration*/
/*	All collected variable are stored	*/
/*	within environement variables		*/
/*						*/
/************************************************/
#include	<stdio.h>
#include	<memory.h>
#include	<errno.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"subapl.h"
#include	"utlcfg.h"


#define	VZCONF	"vzgot_config"
#define	NAMES	"names"

/*

*/
/************************************************/
/*						*/
/*	Procedure to add a file contents to	*/
/*	the env area.				*/
/*						*/
/************************************************/
static int gesenv(char *filename,int toadd)

{
#define	OPEP	"utlcfg.c:gesenv,"

int status;
FILE *fichier;

status=1;
if ((fichier=fopen(filename,"r"))!=(FILE *)0) {
  u_int numline;
  char line[1000];

  status=0;
  numline=0;
  while (apl_getstr(fichier,line,sizeof(line),'#')!=(char *)0) {
    char var[100];
    char contents[200];

    numline++;
    (void) memset(var,'\000',sizeof(var));
    (void) memset(contents,'\000',sizeof(contents));
    if (strlen(line)==0)	/*empty line	*/
      continue;
    if (sscanf(line,"%100[^=]=%199s",var,contents)!=2) {
      (void) apl_alert(0,"%s filename:<%s> line %04d:<%s> invalid format",
			  OPEP,filename,numline,line);
      continue;
      }
    if (contents[0]=='"') {
      char *ptr;

      (void) memmove(contents,contents+1,strlen(contents+1)+1);
      if ((ptr=strchr(contents,'"'))!=(char *)0)
        *ptr='\000';
      } 
    if (toadd==true) {
      char *ptr;

      (void) asprintf(&ptr,"%s=%s",var,contents);
      if (putenv(ptr)!=0) 
        (void) apl_alert(0,"%s unable to add '%s' to env (error=<%s>)",
			    OPEP,ptr,strerror(errno));
      }
    else {
      if (unsetenv(var)<0) 
        (void) apl_alert(0,"%s unable to remove '%s' from env (error=<%s>)",
			    OPEP,var,strerror(errno));
      }
    } 
  (void) fclose(fichier);
  }
else
  (void) apl_alert(0,"%s unable to open file '%s' (error=<%s>)",
		      OPEP,filename,strerror(errno));
return status;
#undef OPEP
}

/*

*/
/************************************************/
/*						*/
/*	loading local variable related to one	*/
/*	container(VPS)				*/
/*	Return 0 if configuation is loaded	*/
/*	1 otherwise.			*/
/*						*/
/************************************************/
int cfg_loadconfig(char *contname)

{
int status;
char *pathetc;
char *conffile;
int phase;
int proceed;

status=0;
pathetc=apl_appdir(d_etc);
conffile=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Loading main config	*/
      (void) asprintf(&conffile,"%s%s",pathetc,VZCONF);
      status=gesenv(conffile,true);
      (void) free(conffile);
      break;
    case 1	:	/*do we have contname	*/
      if (contname==(char *)0) 
	proceed=false;
      break;
    case 2	:	/*now VPS definitions	*/
      (void) asprintf(&conffile,"%s%s/%s",pathetc,NAMES,contname);
      status=gesenv(conffile,true);
      (void) free(conffile);
      break;
    case 3	:	/*next VPS config	*/
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  if (status>0)
    break;
  }
(void) free(pathetc);
return status;
}
